;compile:
;zasm --z80 --dotnames -uwy beast.asm beast.tap
; ================================================================
;   Source for target 'tap'
;   ZASM Tape file for Jupiter ACE - BINARY files
;
;   Basen on TAP file template for ZX Spectrum by kio@little-bat.de
;
;   Copyright (c) McKlaud 2020
;
;   Change log:
;
;   v.0.3 - 7/10/2020 - naming, comments (kio)
;   v.0.2 - 7/10/2020 - spelling corrections and housekeeping
;   v.0.1 - 1/10/2020 - first release
;
; ================================================================
;
; fill byte is 0x00
; #code has an additional argument: the block type (flag byte) for the block.
; Since ZASM 4.2.0 the flag byte in #CODE must be set to NONE for ACE TAPs.
; The assembler calculates and appends checksum byte to each segment.
; Note: If a segment is appended without an explicite address, then the sync
; byte and the checksum byte of the preceding segment are not counted when
; calculating the start address of this segment.
;
; Compile with ZASM v 4.2.x or above
; zasm --z80 --dotnames -uwy source.asm target.tap
;
; --z80       - traget Zilog Z80 (default)
; --dotnames  - allow label names starting with a dot '.' - if needed
; -b          - write output in binary file (default)
; -u          - include object code in list file
; -w          - append label listing to list file
; -y          - include cpu clock cycles in list file
;
; zasm -uwy source.asm target.tap
;
; Load with:
;   0 0 bload binary
;
; Run the code with:
;   16000 call
;
; ================================================================

startadr        equ     $4000;$3E80           ; Start address for BIN files; e.g. 16000

;------------------
; BIN file definition
;------------------
headerlength    equ     25              ; neither block type (flag byte) nor CRC byte included
filetype        equ     $20             ; 0x20 = BINARY file type

;------------------
; Default values of system variables
;------------------
v_c_link        equ     $2020
v_current       equ     $2020
v_context       equ     $2020
v_voclink       equ     $2020
v_stkbot        equ     $2020

;------------------
; ROM routines
;------------------
;CLS             equ     $0A24           ; Clear Screen

#target TAP

#code TAP_HEADER, 0, headerlength, flag=NONE
; Jupiter ACE TAP header structure:

;               defw    headerlength    ; 2 bytes: always 25 bytes (0x1A) for JACE - added by ZASM
                defb    filetype        ; 1 byte:  File Type
                defb    "binary    "    ; 10 bytes: the file name
;                       |----------|     <<< Keep it exactly 10 chars long!
                defw    CODE_DATA_size  ; 2 bytes: File Length
                defw    startadr        ; 2 bytes: Start Address
                defw    v_c_link        ; 2 bytes: current word link (NOT USED)
                defw    v_current       ; 2 bytes: CURRENT (NOT USED)
                defw    v_context       ; 2 bytes: CONTEXT (NOT USED)
                defw    v_voclink       ; 2 bytes: VOCLINK (NOT USED)
                defw    v_stkbot        ; 2 bytes: STKBOT (NOT USED)
;               defb    checksum        ; 1 byte: Header Block CheckSum - added by ZASM


#code CODE_DATA, startadr, *, flag=NONE
;               defw    CODE_DATA_size  ; 2 bytes: TAP 2nd chunck size - added by ZASM

; BIN block starts here
; -----------------------------
;01234
;qwert
;$2400 - vRAM
;$2C00 - chars(write only)

;set font
;0-empty
;1-block
;2-stone
;3-player
;4-beast
	ld hl,fontb
	ld de,$2C00
	ld bc,fonte-fontb
	ldir
gamebeg:
	call fillmap
	call putplayer
;	inc hl
;	ld (hl),1

;stones
	ld bc,$0402
	call puttile

;beasts
	ld bc,$0404
	call puttile

	call scanbeasts
	
;	ld hl,t2
;	ld de,$2408
;	ld bc,15
;	ldir
;	jr $
;init live count
	ld a,7
	ld (pl),a
	ld a,$37
	ld ($2400),a
gamelp:
;----------------------------------------------------------------------------
;                          LOGICAL VIEW OF KEYBOARD
;
;         0     1     2     3     4 -Bits-  4     3     2     1     0
; PORT                                                                    PORT
;
; F7FE  [ 1 ] [ 2 ] [ 3 ] [ 4 ] [ 5 ]  |  [ 6 ] [ 7 ] [ 8 ] [ 9 ] [ 0 ]   EFFE
;  ^                                   |                                   v
; FBFE  [ Q ] [ W ] [ E ] [ R ] [ T ]  |  [ Y ] [ U ] [ I ] [ O ] [ P ]   DFFE
;  ^                                   |                                   v
; FDFE  [ A ] [ S ] [ D ] [ F ] [ G ]  |  [ H ] [ J ] [ K ] [ L ] [ ENT ] BFFE
;  ^                                   |                                   v
; FEFE  [SHI] [SYM] [ Z ] [ X ] [ C ]  |  [ V ] [ B ] [ N ] [ M ] [ SPC ] 7FFE
;  ^            v                                                ^         v
; Start         +------------>--------------------->-------------+        End
;
;
;----------------------------------------------------------------------------
leftkey:
;test for 'R' key
	ld a,$FB
	in a,($FE)
	and 8
	jp z,gamebeg

	ld bc,$fdfe
	ld de,(pxy)
	in a,(c) ; 0-pressed
	bit 0,a
	jp nz,rightk
	call getadr
	push hl
	pop ix
	ld a,(ix-1)
	or a
	jp z,moveleft
;check collision
;2-stone cannot move
	cp 2
	jp z,gamelp
;4-beast - dead
	cp 4
	jp z,death
;1 - moveable block - there are only one
;find space
lfs1:
	dec ix
	ld a,(ix+0)
;check if monster b/w blocks
	cp 4
	jr nz,bb2
	defb $DD,$7E,$FF ; ld a,(ix-1)
	cp 1
	jp nz,gamelp
	call killbeast
	jr lfs2
bb2
	cp 2 ;cannot shift if stone
	jp z,gamelp
	or a
	jp nz,lfs1
;shift left
lfs2:
	ld a,(ix+1)
	ld (ix+0),a
	inc ix
	cp 3 ;until player
	jr nz,lfs2
	ld (ix+0),0
	dec e
 	jp endkeys

moveleft:
	dec e
	ld (ix+0),0
	ld (ix-1),3
 	jp endkeys
rightk:
	bit 2,a
	jp nz,downkey
	call getadr
	push hl
	pop ix
	ld a,(ix+1)
	or a
	jp z,moverig
;check collision

;2-stone cannot move
	cp 2
	jp z,gamelp
;4-beast - dead
	cp 4
	jp z,death
;1 - moveable block - there are only one
;find space
rfs1:
	inc ix
	ld a,(ix+0)
;check if beast b/w blocks
	cp 4
	jr nz,bb1
	ld a,(ix+1)
	cp 1
	jp nz,gamelp
	call killbeast
	jr rfs2
bb1
	cp 2 ;cannot shift if stone
	jp z,gamelp
	or a
	jp nz,rfs1
;shift right
rfs2:
	defb $DD,$7E,$FF ; wrong assembly ld a,(ix-1) FUCK YOU, developer!
	ld (ix+0),a
	dec ix
	cp 3 ;until player
	jr nz,rfs2
	ld (ix+0),0
	inc e
 	jp endkeys

moverig:
	inc e
	ld (ix+0),0
	ld (ix+1),3
 	jp endkeys

downkey:
	bit 1,a
	jp nz,upkey
	call getadr
	push hl
	pop ix
	ld a,(ix+32)
	or a
	jp z,movedown
;check collision
;2-stone cannot move
	cp 2
	jp z,gamelp
;4-beast - dead
	cp 4
	jp z,death
;1 - moveable block - there are only one
;find space
	ld bc,32
dfs1:
	add ix,bc
	ld a,(ix+0)
;check if monster b/w blocks
	cp 4
	jr nz,bb4
	ld a,(ix+32)
	cp 1
	jp nz,gamelp
	call killbeast
	jr dfs22
bb4
	cp 2 ;cannot shift if stone
	jp z,gamelp
	or a
	jr nz,dfs1
;shift down
dfs22:
	ld bc,$FFE0;-32
dfs2:
	defb $DD,$7E,$E0;ld a,(ix-32)
	ld (ix+0),a
	add ix,bc
	cp 3 ;until player
	jr nz,dfs2
	ld (ix+0),0
	inc d
 	jp endkeys

movedown:
	inc d
	ld (ix+0),0
	ld (ix+32),3
 	jp endkeys
 	jp gamelp
upkey:
	ld bc,$fBfe
	in a,(c) ; 0-pressed
	bit 1,a
	jp nz,gamelp
	call getadr
	push hl
	pop ix
	ld a,(ix-32)
	or a
	jp z,moveup
;check collision
;2-stone cannot move
	cp 2
	jp z,gamelp
;4-beast - dead
	cp 4
	jp z,death
;1 - moveable block - there are only one
;find space
	ld bc,$FFE0 ; -32
ufs1:
	add ix,bc
	ld a,(ix+0)
;check if monster b/w blocks
	cp 4
	jr nz,bb3
	defb $DD,$7E,$E0;ld a,(ix-32)
	cp 1
	jp nz,gamelp
	call killbeast
	jr ufs22
bb3
	cp 2 ;cannot shift if stone
	jp z,gamelp
	or a
	jr nz,ufs1
;shift up
ufs22:
	ld bc,32
ufs2:
	defb $DD,$7E,$20;ld a,(ix+32)
	ld (ix+0),a
	add ix,bc
	cp 3 ;until player
	jr nz,ufs2
	ld (ix+0),0
	dec d
 	jp endkeys

moveup:
	dec d
	ld (ix+0),0
	ld (ix-32),3
 	jp endkeys

endkeys:
	ld (pxy),de

wk:
	ld hl,$3c2b
	ld b,9

fr1: ld a,(hl)
fr: cp (hl)
	jr z,fr
	djnz fr1

;check if all beasts dead
	ld b,4
	ld ix,montab
	xor a
chb:
	add a,(ix+0)
	inc ix
	inc ix
	inc ix
	djnz chb
	cp 4
	jr nz,nowin
	ld hl,t2
	ld de,$2408
	ld bc,15
	ldir
	call ws
	jp gamebeg
nowin:
;	call move_beast
;--------------------------------------------
movp equ 1
mbv1 ld a,movp
;mbv2 sub 0
	dec a
	jr nz,ex_mb
	ld ix,montab
	ld b,4
mb_1:
	ld a,(ix+0) ;?dead
	or a
;	call z,bemove
	jr nz,mbnxt

	ld e,(ix+1)
	ld d,(ix+2)
	ld hl,(pxy)
	ld a,l
	sub e
	call nz,sign
	add a,e
	ld e,a
	ld a,h
	sub d
	call nz,sign
	add a,d
	ld d,a
	call getadr
	ld a,(hl)
;must be player
	cp 3
	jr nz,noplayer
	call putbeast
	jp death2
;	jr $
noplayer:
	or a
	jr nz, mbnxt
	call putbeast
mbnxt:
	inc ix
	inc ix
	inc ix
	djnz mb_1
	ld a,movp
ex_mb:
	ld (mbv1+1),a
;--------------------------------------------

;	xor a
;	in a,($FE)
;	cpl
;	and 31
;	jr nz,wk
 	jp gamelp

death:
	ld (ix+0),0
death2:
;------------------------
	xor a
	ld c,200
beep1: ld b,24
	out ($FE),a
	djnz $
	xor 8
;	out ($FE),a	
	dec c
	jr nz,beep1

;------------------------
	ld a,(pl)
	dec a
	ld (pl),a
	add a,$30
	ld ($2400),a
	cp $30
	jr nz,nogameover

	ld hl,t1
	ld de,$2408
	ld bc,16
	ldir
	call ws
	jp gamebeg
nogameover:
	call putplayer
 	jp gamelp
;--------------routines--------------
cls:
	ld hl,$2400
	ld de,$2401
	ld bc,767
	ld (hl),l
	ldir
	ret
random:
	PUSH BC
	LD A,(R1)
	LD C,A
	LD A,(R2)
	ADD A,C
	LD C,A
	LD (R1),A
	LD A,(R3) 
	SUB C
	LD C,A
	LD(R2),A
	RRCA
	LD (R3),A
	LD A,C
	POP BC
	RET
R1      defb $15
R2      defb $70
R3      defb $FD 

rnda: ;random address
	call random
	ld l,a
	call random
	and 3
	add a,$24
	ld h,a
	ld a,(hl)
	or a
	ret
getadr: ;D=y,E=x
	push de
	ld l,d
	add hl,hl ;*2
	add hl,hl ;*4
	add hl,hl ;*8
	ld h,9
	add hl,hl
	add hl,hl
	ld d,0
	add hl,de
	pop de
	ret
fillmap:
	call cls
; draw border
	ld a,2
	ld b,32
	ld hl,$2400
	ld de,$2400+23*32
fm2:
	ld (hl),a
	inc hl
	ld (de),a
	inc de
	djnz fm2
	ld b,32
	ld ix,$2400
	ld de,32
fm3:
	ld (ix+00),a
	ld (ix+31),a
	add ix,de
	djnz fm3
;fill blocks
	ld bc,300;750/2
fm1:
	call rnda
	jr nz,fm1
	ld (hl),1
	dec bc
	ld a,b
	or c
	jr nz,fm1
	

	ret
putplayer:
	ld bc,$0103
	call puttile
	ld (pxy),de
	ret
puttile:
;put tile B=amount,C=value
fm4:
	call random
	and 31
	ld e,a
	call random
	and $17
	ld d,a
	call getadr
	ld a,(hl)
	or a
	jr nz,fm4
	ld (hl),c
	djnz fm4
	ret

scanbeasts:
	ld ix,montab
	ld hl,$2400
	
	ld d,0
sby:
	ld e,0
sbx:
	ld a,(hl)
	cp 4
	jr nz, nextsb
	ld (ix+0),0 ; reset 'dead' flag
	inc ix
	ld (ix+0),e
	inc ix
	ld (ix+0),d
	inc ix
nextsb:
	inc hl
	inc e
	ld a,e
	cp 32
	jr nz,sbx
	inc d
	ld a,d
	cp 24
	jr nz,sby
	ret


;move_beast:
;movp equ 6
;mbv1 ld a,movp
;	dec a
;	jr nz,ex_mb
;	ld ix,montab
;	ld b,4
;mb_1:
;	ld a,(ix+0) ;?dead
;	or a
;	call z,bemove
;mbnxt:
;	inc ix
;	inc ix
;	inc ix
;	djnz mb_1
;	ld a,movp
;ex_mb:
;	ld (mbv1+1),a
;	ret
;bemove:
;	ld e,(ix+1)
;	ld d,(ix+2)
;	ld hl,(pxy)
;	ld a,l
;	sub e
;	call nz,sign
;	add a,e
;	ld e,a
;	ld a,h
;	sub d
;	call nz,sign
;	add a,d
;	ld d,a
;	call getadr
;	ld a,(hl)
;	or a
;	ret nz
;;must be player

putbeast:
	push de
	ld e,(ix+1)
	ld d,(ix+2)
	call getadr
	ld (hl),0
	pop de
	call getadr
	ld (hl),4
	ld (ix+1),e
	ld (ix+2),d
	ret
killbeast:
	push ix
	push de
	push ix
	pop bc
	ld a,4
	ld ix,montab
kb1:
	push af
	ld a,(ix+0)
	or a
	jr nz,kb_nxt

	ld e,(ix+1)
	ld d,(ix+2)
	call getadr
	ld a,c
	cp l
	jr nz,kb_nxt
	ld a,b
	cp h
	jr nz,kb_nxt
	ld (hl),0
	ld (ix+0),1
kb_nxt:
	inc ix
	inc ix
	inc ix
	pop af
	dec a
	jr nz,kb1
	pop de
	pop ix
	ret

sign:
	ld a,1
	ret p
	ld a,$FF
	ret

ws: ld a,$7F ; wait spacebar pressed
	in a,($FE)
	and 1
	jr nz,ws
	ret
t1: defb "G A M E  O V E R"
t2: defb "V I C T O R Y !"
pl: defb 0
pxy: defw 0
fontb:
 defb 000,000,000,000,000,000,000,000;0
 defb 170,085,170,085,170,085,170,085;1
 defb 255,255,255,255,255,255,255,255;2
 defb 024,024,000,255,024,024,036,066;3
 defb 000,063,110,120,096,064,096,063;4
fonte:
montab: ;dead,x,y 4 beasts
; --- END ---
endadr          equ     $

;               defb    checksum       ; 1 byte: CODE_DATA Block CheckSum - added by ZASM
#end                                   ; code blocks END
